/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.anfrage;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;

import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterAbfrageObjekt;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.ParameterInfo.SerializableParameterInfo;
import de.bsvrz.puk.param.lib.ParametrierungsAnfrageTyp;
import de.bsvrz.puk.param.lib.SerializableParameter;
import de.bsvrz.puk.param.param.Parametrierung;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Realisierung einer Klasse zur Beantwortung einer Anfrage zum Auslesen von
 * Parametern.<br>
 * Es erden die Parameterdaten für eine übergebene Parameterspezifiktaion
 * ermittelt, dabei wird prinzipiell der Mechanismus der Parametersuche nach
 * lokalen Parametern, Hierarchieparametern und Typ-Parametern mit den
 * zugeordneten Standardparametern durchlaufen. Zusätzlich zur Bestimmung der
 * Parameter für konkrete Systemobjekte über die übliche Schnittstelle der
 * Anmeldung als <code>ClientReceiverInterface</code> für Parameter-Soll-Daten
 * bietet diese Schnittstelle die Möglichkeit Parameter für übergeordnete
 * Objekte innerhalb der Objekthierarchie und Parameter, die an Typobjekten
 * gespeichert wurden zu ermitteln.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class ParameterLesenErmittler extends ParameterAnfrageBeantworter {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** die Parameterdaten der Anfrage. */
	private final ParameterInfo[] anfrageInfos;

	/**
	 * markiert, ob in der Antwort die Daten mit versendet werden solen oder
	 * nicht.
	 */
	private boolean datenLiefern;

	/**
	 * Konstruktor, erzeugt eine Instanz der Anfrage.<br>
	 * Die Anfrage wird mit den Daten des übergebenen Parametrierungsmoduls
	 * bearbeitet. Das ermittelte Ergebnis, wird an das übergebene
	 * Anfrager-Objekt zurückgesendet, wobei das Objekt selbst vom Typ
	 * "Applikation" sein sollte. Der übergebene Index ist Bestandteil der
	 * Antwort und dient auf Client-Seite der Zuordnung von Anfragen und
	 * Antworten. Die Daten enthalten die Anfrage als serialisiertes Array von
	 * Parameterspezifikationen, für die eine Anfrage erfolgen soll.
	 *
	 * @param parametrierung
	 *            das verwendete Parametrierungsmodul
	 * @param anfrager
	 *            das Systemobjekt, das die anfragende Applikation repräsentiert
	 * @param index
	 *            der Index der Anfrage
	 * @param abfrageObjekt
	 *            die Parameterdaten der Anfrage
	 */
	ParameterLesenErmittler(final Parametrierung parametrierung, final SystemObject anfrager, final int index,
			final Object abfrageObjekt) {
		super(parametrierung, anfrager, index);

		SerializableParameterInfo[] daten = null;
		if (abfrageObjekt instanceof SerializableParameterInfo[]) {
			daten = (SerializableParameterInfo[]) abfrageObjekt;
			datenLiefern = true;
		} else if (abfrageObjekt instanceof ParameterAbfrageObjekt) {
			daten = ((ParameterAbfrageObjekt) abfrageObjekt).getParameterInfos();
			datenLiefern = ((ParameterAbfrageObjekt) abfrageObjekt).isDatenLiefern();
		} else {
			throw new IllegalArgumentException(
					"Die Daten der Anfrage konnten nicht interpretiert werden: " + Arrays.toString(daten));
		}

		final SerializableParameterInfo[] anfrageFeld = daten;
		anfrageInfos = new ParameterInfo[daten.length];

		for (int idx = 0; idx < anfrageFeld.length; idx++) {
			anfrageInfos[idx] = anfrageFeld[idx].erzeugeParameterInfo(parametrierung.getVerbindung());
		}
	}

	/**
	 * für die Liste der bei der Anfrage übermittelten Parameteridentifikationen
	 * werden die entsprechenden Datensätze ermittelt. Jede
	 * Parameterspezifikation wird bearbeitet und für jede Spezifikation wird
	 * eine Antwort in das Ergebnis-Feld eingetragen unabhängig davon, ob ein
	 * Parameterdatensatz ermittelt werden konnte oder nicht. Das heißt das
	 * Ergebnis ist ein Feld entsprechend der Größe des Liste der Anfragen und
	 * enthält serialisierte Parameterdatensätze mit gültigen Daten oder mit
	 * leeren Parameterdatensätzen.
	 */
	@Override
	protected Serializable ermittleAntwort() {

		ParameterLesenErmittler.LOGGER.fine("Parameter lesen für", getAnfrager());

		final Collection<SerializableParameter> ergebnis = new ArrayList<>();
		for (final ParameterInfo info : anfrageInfos) {
			final Parameter parameter = getParametrierung().ermittleParameter(info, datenLiefern);
			if (parameter == null) {
				ergebnis.add(new Parameter(info).getSerializableParameter(getParametrierung().getVerbindung()));
			} else {
				ergebnis.add(parameter.getSerializableParameter(getParametrierung().getVerbindung()));
			}
		}

		ParameterLesenErmittler.LOGGER.fine("Es wurden " + ergebnis.size() + " Parameterdatensätze ausgelesen");
		ParameterLesenErmittler.LOGGER.finer("Ermittelte Parameter:", ergebnis);

		final SerializableParameter[] ergebnisFeld = ergebnis.toArray(new SerializableParameter[ergebnis.size()]);

		return ergebnisFeld;
	}

	@Override
	protected ParametrierungsAnfrageTyp getAntwortTyp() {
		return ParametrierungsAnfrageTyp.PARAMETER;
	}
}
