/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.daten.DataWithTime;
import de.bsvrz.puk.param.lib.daten.Datensatz;
import de.bsvrz.puk.param.lib.daten.ParameterDatenFactory;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.bitctrl.io.FileTool;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Persistenz, welche Daten in einem Verzeichnis ablegt, wobei die Speicherung
 * nach PIDs erfolgt.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class FilePersistenzPid implements Persistenz {

	/** der Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** das Basisverzeichnis zur Ablage der Daten. */
	private File verzeichnis;

	/** die verwendete Datenverteilerverbindung. */
	private ClientDavInterface verbindung;

	/**
	 * Konstruktor, erzeugt eine Instanz des Persistenzmoduls.
	 */
	FilePersistenzPid() {
		FilePersistenzPid.LOGGER.info("Starte Submodul Persistenz PID ...");
	}

	@Override
	public void aufraeumen() {
		// es sind keine Operationen vorgesehen
	}

	@Override
	public void close() {
		FilePersistenzPid.LOGGER.info("Persistenzmodul wird beendet");
	}

	@Override
	public void entferne(final ParameterInfo parameter) {
		final List<Map<String, Serializable>> neueMenge = new ArrayList<>();

		final File datei = getParameterDatei(parameter, false);
		if (datei != null) {

			boolean geloescht = false;

			final List<Map<String, Serializable>> menge = getParameterMenge(datei);

			for (final Map<String, Serializable> i : menge) {
				boolean behalten = true;
				if (i.get("Obj").equals(parameter.getObjekt().getId())
						&& i.get("Typ").equals(parameter.getTyp().getId())
						&& i.get("Atg").equals(parameter.getAtg().getId()) && i.get("Sim").equals(parameter.getSim())) {
					behalten = false;
				}

				if (behalten) {
					neueMenge.add(i);
				} else {
					geloescht = true;
				}
			}

			if (geloescht) {
				try (ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(datei))) {
					out.writeObject(neueMenge);
				} catch (final FileNotFoundException e) {
					throw new IllegalStateException(
							"Datei: " + datei.getAbsolutePath() + " konnte nicht gefunden werden", e);
				} catch (final IOException e) {
					throw new IllegalStateException("Fehler beim Schreiben der Datei: " + datei.getAbsolutePath(), e);
				}
			}
		}
	}

	@Override
	public void entferneSimulation(final short simulationsVariante) {
		if (simulationsVariante > 0) {
			final File simVerzeichnis = new File(verzeichnis, Short.toString(simulationsVariante));
			if (!FileTool.loescheVerzeichnis(simVerzeichnis)) {
				LogTools.log(FilePersistenzPid.LOGGER, ParamMessage.ErrorEntfernenVerzeichnis,
						simVerzeichnis.getAbsolutePath());
			}
		}
	}

	@Override
	public void erzeugeSimulation(final short simulationsVariante) {
		if (simulationsVariante > 0) {
			final File basisVerzeichnis = new File(verzeichnis, "0");
			final File simVerzeichnis = new File(verzeichnis, Short.toString(simulationsVariante));
			if (!FileTool.kopiereVerzeichnis(basisVerzeichnis, simVerzeichnis)) {
				LogTools.log(FilePersistenzPid.LOGGER, ParamMessage.ErrorAnlegenSimulationsVerzeichnis,
						simVerzeichnis.getAbsolutePath());
			}
		}
	}

	/**
	 * liefert eine konstante ID, die dem String zugeordnet werden kann. Die
	 * Funktion entspricht der Bildung des Hashcodes der Klasse String in Java
	 * 1.6.
	 *
	 * @param pid
	 *            die PID
	 * @return der Code
	 */
	private int getKonstanteId(final String pid) {
		int id = 0;
		if (pid != null) {
			final char[] val = pid.toCharArray();
			for (int i = 0; i < val.length; i++) {
				id = 31 * id + val[i];
			}
		}
		return id;
	}

	/**
	 * ermittelt die Datei für die Ablage der jeweiligen Parameterdaten.
	 *
	 * @param parameter
	 *            die Parameterspezifikation
	 * @param anlegen
	 *            Datei bei Bedarf anlegen ?
	 * @return die Datei
	 */
	private File getParameterDatei(final ParameterInfo parameter, final boolean anlegen) {

		/* Parameterdatei */
		final File datei = getPidFile(verzeichnis, parameter.getObjekt().getPid(), parameter.getSim(), anlegen);
		// new File(verzeichnis.getAbsolutePath(), id + ".param");

		// Wenn die Parameterdatei noch nicht existiert, wird sie bei Bedarf mit
		// einer leeren Parameterliste initialisiert
		if (!datei.exists() && anlegen) {

			/* Parametermenge */
			final List<Map<String, Serializable>> menge = new ArrayList<>();

			try (ObjectOutputStream ausgabe = new ObjectOutputStream(new FileOutputStream(datei))) {
				ausgabe.writeObject(menge);
			} catch (final IOException e) {
				throw new IllegalStateException(
						"Die Datei " + datei.getAbsolutePath() + " konnte nicht geschrieben werden", e);
			}
		}

		return datei;
	}

	/**
	 * liest die Parameterobjekte aus einer Datei ein.
	 *
	 * @param datei
	 *            die einzulesende Datei
	 * @return die Liste der Parametereinträge
	 */
	@SuppressWarnings("unchecked")
	private List<Map<String, Serializable>> getParameterMenge(final File datei) {

		/* Parametermenge */
		final List<Map<String, Serializable>> menge = new ArrayList<>();

		try (ObjectInputStream eingabe = new ObjectInputStream(new FileInputStream(datei))) {
			final Object obj = eingabe.readObject();
			menge.addAll((List<Map<String, Serializable>>) obj);
			eingabe.close();
		} catch (final ClassNotFoundException e) {
			throw new IllegalStateException("Die gespeicherten Daten der Datei " + datei.getAbsolutePath()
					+ " entsprechen nicht der erwarteten Form", e);
		} catch (final IOException e) {
			if (datei.exists()) {
				if (!datei.delete()) {
					throw new IllegalStateException(
							"Die Datei " + datei.getAbsolutePath() + " konnte nicht gelesen werden", e);
				}
				LogTools.log(FilePersistenzPid.LOGGER, ParamMessage.ErrorDefekteParameterDatei,
						datei.getAbsolutePath());
			}
		}

		return menge;
	}

	/**
	 * liefert eine Datei, deren Pfad sich aus der übergebenen Objekt-PID und
	 * der Simulationsvariante ergibt.
	 *
	 * @param basis
	 *            das Basisverzeichnis
	 * @param pid
	 *            die PID des Objekts, dessen Parameter gespeichert werden
	 *            sollen
	 * @param sim
	 *            die Simulationsvariante
	 * @param anlegen
	 *            Verzeichnis anlegen, falls es nicht existiert ?
	 * @return die Datei
	 */
	private File getPidFile(final File basis, final String pid, final short sim, final boolean anlegen) {

		final StringBuffer name = new StringBuffer();
		name.append(getKonstanteId(pid));
		File result = new File(basis, Short.toString(sim));
		while (name.length() > 0) {
			result = new File(result, name.substring(0, Math.min(name.length(), 3)));
			name.replace(0, Math.min(name.length(), 3), "");
		}

		if (anlegen) {
			if (!result.exists()) {
				result.mkdirs();
			}
		}

		name.append(pidAlsDateiName(pid));
		while (name.length() > 15) {
			result = new File(result, name.substring(0, 15));
			name.replace(0, 15, "");
		}

		if (anlegen) {
			if (!result.exists()) {
				result.mkdirs();
			}
		}

		name.append(".par");
		result = new File(result, name.toString());

		return result;
	}

	@Override
	public void initalisiere(final ClientDavInterface davInterface, final PersistenzParameter parameter) {
		this.verbindung = davInterface;
		String name = parameter.getPersistenzWurzel();
		if (name == null || name.length() <= 0) {
			name = "parameter";
		}
		final File file = new File(name);
		if (!file.exists()) {
			FilePersistenzPid.LOGGER.info("Lege Parameterverzeichnis an: " + file.getAbsolutePath());
			file.mkdir();
		}
		if (!file.isDirectory()) {
			throw new IllegalArgumentException("Es wurde kein gültiger Verzeichnispfad angegeben.");
		} else if (!file.canRead()) {
			throw new IllegalArgumentException("Das angegebenen Verzeichnis ist nicht lesbar.");
		} else if (!file.canWrite()) {
			throw new IllegalArgumentException("Das angegebenen Verzeichnis ist nicht schreibbar.");
		}

		this.verzeichnis = file;
	}

	@Override
	public Parameter lies(final ParameterInfo parameter) {
		final Parameter result = new Parameter(parameter);
		final List<Map<String, Serializable>> menge = getParameterMenge(getParameterDatei(parameter, false));
		for (final Map<String, Serializable> i : menge) {
			if (i.get("Obj").equals(result.getObjekt().getId()) && i.get("Typ").equals(result.getTyp().getId())
					&& i.get("Atg").equals(result.getAtg().getId()) && i.get("Sim").equals(result.getSim())) {
				result.setDataWithTime(new DataWithTime(
						ParameterDatenFactory.getData(verbindung, result.getAtg(), (Datensatz) i.get("Datenblock")),
						System.currentTimeMillis()));
			}
		}

		FilePersistenzPid.LOGGER.finer("Daten gelesen", new Object[] { parameter, result.getData() });
		return result;
	}

	/**
	 * konvertiert eine PID in einen im Dateisystem speicherbaren Name.
	 *
	 * @param pid
	 *            die PID
	 * @return der Name
	 */
	private String pidAlsDateiName(final String pid) {
		final StringBuffer result = new StringBuffer();

		for (final Character ch : pid.toCharArray()) {
			if (ch.charValue() < 127) {
				result.append(ch);
			} else {
				result.append('#');
				result.append(Long.toString(ch.charValue()));
				result.append('#');
			}
		}

		return result.toString();
	}

	@Override
	public void schreibe(final Parameter parameter) {

		/* zu schreibender Parametersatz */
		final Map<String, Serializable> satz = new HashMap<>();

		/* Parameterdatei für das Systemobjekt */
		final File datei = getParameterDatei(parameter.getInfo(), true);

		/* gesicherte Parametermenge */
		final List<Map<String, Serializable>> menge = getParameterMenge(datei);

		// Parametersatz zusammenstellen
		satz.put("Obj", parameter.getObjekt().getId());
		satz.put("Typ", parameter.getTyp().getId());
		satz.put("Atg", parameter.getAtg().getId());
		satz.put("Sim", parameter.getSim());
		satz.put("Datenblock", ParameterDatenFactory.getSerializable(verbindung, parameter.getData()));

		// Den vorhandene Parametersatz ändern ...
		boolean gefunden = false;
		for (final Map<String, Serializable> i : menge) {
			if (i.get("Obj").equals(satz.get("Obj")) && i.get("Typ").equals(satz.get("Typ"))
					&& i.get("Atg").equals(satz.get("Atg")) && i.get("Sim").equals(satz.get("Sim"))) {
				i.put("Datenblock", satz.get("Datenblock"));
				gefunden = true;
				break;
			}
		}

		// ... oder wenn nicht vorhanden, den Parametersatz hinzufügen
		if (!gefunden) {
			menge.add(satz);
		}

		// Parametermenge zurück in Datei schreiben
		try (ObjectOutputStream out = new ObjectOutputStream(new FileOutputStream(datei))) {
			out.writeObject(menge);
		} catch (final FileNotFoundException e) {
			throw new IllegalStateException("Datei: " + datei.getAbsolutePath() + " konnte nicht gefunden werden", e);
		} catch (final IOException e) {
			throw new IllegalStateException("Fehler beim Schreiben der Datei: " + datei.getAbsolutePath(), e);
		}

		FilePersistenzPid.LOGGER.finer("Daten geschrieben", new Object[] { parameter, datei.getAbsolutePath() });
	}

	@Override
	public void backup() {
		FilePersistenzPid.LOGGER.error("Backup ist für die File-Persistenz nicht implementiert");
	}

	@Override
	public void cleanup(final ParameterProvider provider) {
		FilePersistenzPid.LOGGER.error("Cleanup ist für die File-Persistenz nicht implementiert");
	}
}
