/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import java.lang.Thread.UncaughtExceptionHandler;
import java.text.MessageFormat;

import de.bsvrz.dav.daf.main.ApplicationCloseActionHandler;
import de.bsvrz.dav.daf.main.ClientDavConnection;
import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.ClientDavParameters;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList;
import de.bsvrz.sys.funclib.commandLineArgs.ArgumentList.Argument;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.sys.funclib.operatingMessage.MessageSender;

/**
 * Die Applikation "Parametrierung": SE8, SWE8.2.<br>
 * Die Klasse realisiert das Modul "Kern" entsprechend der Feinspezifikation.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
final class ParamApp implements ApplicationCloseActionHandler, UncaughtExceptionHandler {

	/**
	 * Private Klasse zur Verwaltung der Ready-Message der Applikation.
	 */
	static class ReadySetter {

		/** der Status der Fertigmeldung. */
		private boolean ready;

		/**
		 * versendet die Fertigmeldung der Applikation, wenn si nicht bereits
		 * versendet wurde.
		 *
		 * @param connection
		 *            die Verbindung über die die Meldung versendet wird.
		 */
		void setReady(final ClientDavInterface connection) {
			if (!ready) {
				if (connection != null) {
					connection.sendApplicationReadyMessage();
					ready = true;
				}
			}
		}
	}

	/**
	 * Handler der beim Beenden der Applikation ausgeführt wird.
	 */
	private class ShutDownHandler extends Thread {

		/** Logger für Debugausgaben. */
		private final Debug shutdownLogger = Debug.getLogger();

		/** die zugeordnete Parametrierungsapplikation. */
		private final ParamApp paramApp;

		/**
		 * Konstruktor.
		 *
		 * @param paramApp
		 *            die zugeordnete Applikation
		 */
		ShutDownHandler(final ParamApp paramApp) {
			this.paramApp = paramApp;
		}

		@Override
		public void run() {
			shutdownLogger.info("Beende Simulationen");
			paramApp.beendeSimulationen();
			shutdownLogger.info("Parametrierung wurde beendet");
		}
	}

	/** der Name des Standard-Persistenz-Moduls. */
	private static final String DEFAULT_PERSISTENZ_MODUL = DerbyPersistenz.class.getName();

	/**
	 * der Argumentname zur Angabe des Klassennamens für das Persistenzmodul.
	 */
	private static final String PARAMETERNAME_PERSISTENZ_MODUL = "-persistenzModul";

	/**
	 * der Argumentname zur Angabe der PID des Objekts, das die Parameter der
	 * Parametrierung speichert.
	 */
	private static final String PARAMETERNAME_PARAMETRIERUNG = "-parametrierung";

	/**
	 * bestimmt, ob Standardparameter mit dem "alten" Aspekt
	 * "asp.parameterDefault" ausgelesen werden sollen.
	 */
	private static final String PARAMETERNAME_OLD_DEFAULT = "-oldDefault";

	/**
	 * bestimmt, ob die Parametrierung eine Fertigmeldung erst abgibt, wenn der
	 * SettingsManager das erste Mal den Parameterdatensatz analysiert hatte.
	 */
	private static final String PARAMETERNAME_WAIT_FOR_SETTINGSMANAGER = "-waitForSettingsManager";

	/**
	 * Hauptfunktion der Applikation.<br>
	 * Es wird eine Datenverteiler-Standardapplikation installiert.
	 *
	 * @param argumente
	 *            die Kommandozeilen-Argumente
	 */
	public static void main(final String[] argumente) {

		final ParamApp application = new ParamApp();
		final StringBuilder applicationLabel = new StringBuilder();
		for (final String arg : argumente) {
			applicationLabel.append(arg);
		}
		final ArgumentList argumentList = new ArgumentList(argumente);
		final String[] classNameParts = application.getClass().getName().split("[.]");
		final int lastPartIndex = classNameParts.length - 1;
		String applicationName = null;
		if (lastPartIndex < 0) {
			applicationName = "StandardApplication";
		} else {
			applicationName = classNameParts[lastPartIndex];
		}
		applicationLabel.append(applicationName);
		application.initLogger(applicationName, argumentList);

		try {
			// ArgumentListe wird in ClientDavParameters konvertiert
			final ClientDavParameters parameters = new ClientDavParameters(argumentList);
			parameters.setApplicationTypePid("typ.parametrierungsApplikation");
			parameters.setApplicationName(applicationName);
			// zuerst darf die Applikation die ArgumentListe durcharbeiten
			application.parseArguments(argumentList);
			argumentList.ensureAllArgumentsUsed();
			final ClientDavInterface connection = new ClientDavConnection(parameters);

			// Fertigmeldung für Start/Stop wird eplizit selbst übernommen
			connection.enableExplicitApplicationReadyMessage();

			connection.connect();
			connection.login();

			/* ApplikationsKennung */
			applicationLabel.append(connection.getLocalConfigurationAuthority().getPid());

			// MessageSender funktioniert hier nicht, da die Parametrierung die
			// Nutzer-ID 0 erhält, was der Messagesender als "nicht angemeldet"
			// interpretiert und daraufhin eine RuntimeException wirft.
			// MessageSender.getInstance().init(connection, applicationName,

			application.initialize(connection);

			// Fertigmeldung wird gesendet, wenn nicht auf den SettingsManager
			// gewartet werden soll.
			if (!application.waitForSettingsManager) {
				connection.sendApplicationReadyMessage();
			}

		} catch (final Exception ex) {
			application.logger.error("Fehler", ex);
			System.exit(1);
		}
	}

	private void initLogger(String applicationName, ArgumentList argumentList) {
		Debug.init(applicationName, argumentList);
		logger = Debug.getLogger();
	}

	/** der Ausgabekanal für Logging-Ausgaben. */
	private Debug logger;

	/** die PID des Objekts, das die Parameter der Parametrierung speichert. */
	private String parametrierungsObjektPid;

	/** der Klassenname des Persistenzmoduls. */
	private String persistenzModul = ParamApp.DEFAULT_PERSISTENZ_MODUL;

	/**
	 * markiert, ob Standardparameter mit dem "alten" Aspekt
	 * "asp.parameterDefault" ausgelesen werden sollen.
	 */
	private boolean oldDefault;

	/** das verwendete Verwaltungsmodul. */
	private Verwaltung verwaltung;

	/** die verwendete Datenverteilerverbindung. */
	private ClientDavInterface davVerbindung;

	/**
	 * markiert, ob eine Fertigmeldung der Parametrierung erst erfolgt, wenn der
	 * SettingsManager den Parameterdatensatz erstmalig analysiert hat.
	 */
	private boolean waitForSettingsManager = true;

	/** ein Objekt zur Verwaltung der Fertigmeldung der Applikation. */
	private final ReadySetter readySetter = new ReadySetter();

	/**
	 * Standardkonstruktor.
	 * @param argumente 
	 */
	private ParamApp() {
		// es gibt keine externen Instanzen der Klasse
	}

	/** beendet alle aktiven Simulationen. */
	void beendeSimulationen() {
		if (verwaltung != null && verwaltung.getSimulationsVerwaltung() != null) {
			verwaltung.getSimulationsVerwaltung().beendeSimulationen();
		}
		if (davVerbindung != null) {
			davVerbindung.sleep(10000);
		}
	}

	@Override
	public void close(final String error) {
		if( error == null || error.isEmpty()) {
			logger.info("Parameter-Applikation wurde beendet");
		} else {
			logger.info("Parameter-Applikation wurde beendet: " + error);
		}
		if (verwaltung != null) {
			verwaltung.close();
		}
		System.exit(0);
	}

	/**
	 * liefert die verwendete Datenverteilerverbindung.
	 *
	 * @return die Verbindung
	 */
	public ClientDavInterface getDavVerbindung() {
		return davVerbindung;
	}

	private void initialize(final ClientDavInterface verbindung) {
		davVerbindung = verbindung;
		verbindung.setCloseHandler(this);

		MessageSender.getInstance()
				.setApplicationLabel(getClass().getSimpleName() + "_" + verbindung.getLocalApplicationObject().getId());

		if (verbindung.getClientDavParameters().getSimulationVariant() != 0) {
			Debug.getLogger().info("Die Parametrierung kann nicht mit Simulationsvariante \""
					+ verbindung.getClientDavParameters().getSimulationVariant() + "\" gestartet werden.");
			System.exit(-1);
		}

		try {
			logger.info("Erzeuge Persistenzmodul: " + persistenzModul + " in Verzeichnis "
					+ PersistenzParameter.getInstanz().getPersistenzWurzel());
			final Persistenz persistenz = (Persistenz) Class.forName(persistenzModul).newInstance();
			persistenz.initalisiere(verbindung, PersistenzParameter.getInstanz());

			logger.info("Ermittle Parametrierungsobjekt -- PID = " + parametrierungsObjektPid);
			SystemObject parametrierungsObjekt = null;
			if (parametrierungsObjektPid != null && parametrierungsObjektPid.length() > 0) {
				parametrierungsObjekt = verbindung.getDataModel().getObject(parametrierungsObjektPid);
			} else {
				parametrierungsObjekt = verbindung.getLocalConfigurationAuthority();
			}

			if (parametrierungsObjekt == null) {
				throw new RuntimeException(
						"Es konnte kein Objekt zur Bereitstellung der Parametrierungsparameter ermittelt werden.");
			}

			if (!parametrierungsObjekt.isOfType(ParaKonstanten.PID_TYP_PARAMETRIERUNG)) {
				throw new RuntimeException("Das Parametrierungsobjekt " + parametrierungsObjekt.getNameOrPidOrId()
						+ " ist nicht vom Typ Parametrierung.");
			}

			verwaltung = new Verwaltung(this, persistenz, parametrierungsObjekt, oldDefault);
		} catch (final Exception e) {
			throw new RuntimeException("Das Persistenzobjekt konnte nicht angelegt werden: " + e.getMessage(), e);
		}
	}

	private void parseArguments(final ArgumentList argumente) {

		Runtime.getRuntime().addShutdownHook(new ShutDownHandler(this));
		Thread.setDefaultUncaughtExceptionHandler(this);

		persistenzModul = argumente
				.fetchArgument(ParamApp.PARAMETERNAME_PERSISTENZ_MODUL + "=" + ParamApp.DEFAULT_PERSISTENZ_MODUL)
				.asNonEmptyString();
		parametrierungsObjektPid = argumente.fetchArgument(ParamApp.PARAMETERNAME_PARAMETRIERUNG + "=").asString();
		PersistenzParameter.getInstanz()
				.setzeParameter(argumente.fetchArgument(PersistenzParameter.PARAMETERNAME_PERSISTENZ));
		if (argumente.hasArgument(PersistenzParameter.PARAMETERNAME_CACHE_ANZAHL)) {
			PersistenzParameter.getInstanz()
					.setzeParameter(argumente.fetchArgument(PersistenzParameter.PARAMETERNAME_CACHE_ANZAHL));
		}
		if (argumente.hasArgument(ParamApp.PARAMETERNAME_OLD_DEFAULT)) {
			final Argument arg = argumente.fetchArgument(ParamApp.PARAMETERNAME_OLD_DEFAULT);
			oldDefault = arg.booleanValue();
		}

		if (argumente.hasArgument(ParamApp.PARAMETERNAME_WAIT_FOR_SETTINGSMANAGER)) {
			final Argument arg = argumente.fetchArgument(ParamApp.PARAMETERNAME_WAIT_FOR_SETTINGSMANAGER);
			waitForSettingsManager = arg.booleanValue();
		}

		if (argumente.hasArgument(PersistenzParameter.PARAMETERNAME_BEREINIGUNG_ZYKLUS)) {
			PersistenzParameter.getInstanz()
					.setzeParameter(argumente.fetchArgument(PersistenzParameter.PARAMETERNAME_BEREINIGUNG_ZYKLUS));
		}
	}

	/**
	 * sendet die Fertigmeldung der Applikation über die definierte
	 * Datenverteilerverbindung.
	 */
	public void sendReadyMessage() {
		readySetter.setReady(davVerbindung);
	}

	@Override
	public void uncaughtException(final Thread t, final Throwable e) {
		logger.error(MessageFormat.format(ParamMessage.ErrorParamAbbruch.toString(), t, e ), e);
		System.exit(-1);
	}
}
