/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.puk.param.hierarchie.HierarchieObjekt;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.QuellTyp;
import de.bsvrz.puk.param.lib.daten.DataWithTime;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * die Klasse realisiert den Suchalgorithmus zum Bestimmen der Daten einer
 * Parameteridentifikation.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class ParameterErmittler {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die zu Grunde liegende Parametrierung, die die Daten für die
	 * Parameterdatensätze zur Verfügung stellt.
	 */
	private final Parametrierung parametrierung;

	/**
	 * erzeugt eine Instanz des Ermittlers für die übergebene Parametrierung.
	 *
	 * @param parametrierung
	 *            die Parametrierung
	 */
	protected ParameterErmittler(final Parametrierung parametrierung) {
		this.parametrierung = parametrierung;
	}

	/**
	 * ermittelt den innerhalb der Persistenz der Parametrierung gespeicherten
	 * Parameter für die gegebene Parameterspezifikation.
	 *
	 * @param info
	 *            die Spezifikation des gesuchten Parameters
	 * @return der ermittelte Parameter
	 */
	private Parameter ermittleLokalenParameter(final ParameterInfo info) {

		final Parameter result = parametrierung.getPersistenz().lies(info);

		if (result.getData() != null) {
			if (result.getData().isDefined()) {
				result.setQuelle(info.getObjekt(), QuellTyp.LOKAL_OBJEKT);
			} else {
				LogTools.log(ParameterErmittler.LOGGER, ParamMessage.WarningUnvollstaendigerDatensatz, info,
						result.getData());
				result.setDataWithTime((DataWithTime) null);
			}
		}

		// gegebenenfalls HierarchieObjekte durchsuchen
		if (result.getData() == null) {
			final HierarchieObjekt objekt = parametrierung.getHierarchieManager()
					.getHierarchieObjekt(info.getObjekt().getId());
			if (objekt != null) {
				Collection<HierarchieObjekt> suchListe = new ArrayList<>();
				suchListe.add(objekt);
				do {
					suchListe = expandiereSuchliste(suchListe);
					for (final HierarchieObjekt eintrag : suchListe) {
						final ParameterInfo hierarchieInfo = new ParameterInfo(eintrag.getObjekt(), info.getTyp(),
								info.getAtg(), info.getSim());
						final Parameter parameter = parametrierung.getPersistenz().lies(hierarchieInfo);
						if (parameter.getData() != null) {
							if (parameter.getData().isDefined()) {
								result.setQuelle(eintrag.getObjekt(), QuellTyp.LOKAL_HIERARCHIE);
								result.setDataWithTime(parameter.getDataWithTime());
								break;
							}
							LogTools.log(ParameterErmittler.LOGGER, ParamMessage.WarningUnvollstaendigerDatensatz,
									hierarchieInfo, result.getData());
							result.setDataWithTime((DataWithTime) null);
						}
					}

				} while (result.getData() == null && suchListe.size() > 0);
			}
		}

		// gegebenenfalls Typdefinitionen durchsuchen
		if (result.getData() == null) {
			SystemObjectType type = info.getObjekt().getType();
			while (result.getData() == null && type != null) {
				final ParameterInfo typInfo = new ParameterInfo(type, info.getTyp(), info.getAtg(), info.getSim());
				final Parameter parameter = parametrierung.getPersistenz().lies(typInfo);
				if (parameter.getData() != null) {
					if (parameter.getData().isDefined()) {
						result.setQuelle(type, QuellTyp.LOKAL_TYP);
						result.setDataWithTime(parameter.getDataWithTime());
					} else {
						LogTools.log(ParameterErmittler.LOGGER, ParamMessage.WarningUnvollstaendigerDatensatz, typInfo,
								result.getData());
						result.setDataWithTime((DataWithTime) null);
					}
				}
				type = getValidSuperType(type, info.getAtg());
			}
		}

		return result;
	}

	/**
	 * ermittelt den Parameter entsprechend der Suchvorgaben.<br>
	 * <br>
	 * <ol>
	 * <li>Von der Parametrierung gespeicherter Parameterdatensatz für das
	 * konkrete Objekt</li>
	 * <li>Von der Parametrierung gespeicherter Parameterdatensatz für ein in
	 * der Hierarchie übergeordnetes Objekt, wobei die Hierarchie komplett
	 * durchlaufen wird</li>
	 * <li>Von der Parametrierung gespeicherter Parameterdatensatz für den Typ
	 * des Objekts</li>
	 * <li>Von der Parametrierung gespeicherter Parameterdatensatz für
	 * Basistypen des Typs des Objekts</li>
	 * <li>Von der Parametrierung gespeicherter Parameterdatensatz für den Typ
	 * eines in der Hierarchie übergeordnetes Objekt, sowie seiner Basistypen
	 * </li>
	 * <li>der konfigurierte Standarddatensatz für das konkrete Objekt</li>
	 * <li>der konfigurierte Standarddatensatz an einem in der Hierarchie
	 * übergeordneten Objekt</li>
	 * <li>der konfigurierte Standarddatensatz am Typ des Objekts</li>
	 * <li>der konfigurierte Standarddatensatz an Basistypen des Typs des
	 * Objekts</li>
	 * <li>der konfigurierte Standarddatensatz für den Typ eines in der
	 * Hierarchie übergeordnetes Objekt, sowie seiner Basistypen
	 * </ol>
	 *
	 * @param info
	 *            die Spezifikation der gesuchten Daten
	 * @return den ermittelten Parameterdatensatz
	 */
	Parameter ermittleParameter(final ParameterInfo info) {

		Parameter result = null;

		final SystemObject objekt = info.getObjekt();
		if (parametrierung.getProvider().isProvided(info) || objekt instanceof SystemObjectType) {
			result = ermittleLokalenParameter(info);
			if (result == null || result.getData() == null) {
				ParameterErmittler.LOGGER
						.finest("Keine Lokalen Parameter für " + info + "verfügbar, verwende Standardparameter!");
				result = ermittleStandardParameter(info);
			} else if (!result.getData().isDefined()) {
				ParameterErmittler.LOGGER.warning("Die Parameterdaten konnten nicht deserialisiert werden!", result);
				result = ermittleStandardParameter(info);
			}
		}

		if (result == null || result.getData() == null || !result.getData().isDefined()) {
			ParameterErmittler.LOGGER.warning("Es konnten keine Parameterdaten für " + info + " ermittelt werden!",
					result);
			result = new Parameter(info);
		}

		return result;
	}

	/**
	 * ermittelt den Standardparameter für die gegebene Parameterspezifikation.
	 *
	 * @param info
	 *            die Spezifikation des gesuchten Parameters
	 * @return der ermittelte Parameter
	 */
	private Parameter ermittleStandardParameter(final ParameterInfo info) {

		final Parameter parameter = new Parameter(info);
		parameter.setDataWithTime(parametrierung.getDefaultParameterDaten(info));

		if (parameter.getData() != null) {
			parameter.setQuelle(info.getObjekt(), QuellTyp.DEFAULT_OBJEKT);
		}

		final Set<SystemObject> hierarchieTypListe = new HashSet<>();

		// gegebenenfalls HierarchieObjekte durchsuchen
		if (parameter.getData() == null) {
			final HierarchieObjekt objekt = parametrierung.getHierarchieManager()
					.getHierarchieObjekt(info.getObjekt().getId());
			if (objekt != null) {
				Collection<HierarchieObjekt> suchListe = new ArrayList<>();
				suchListe.add(objekt);
				do {
					suchListe = expandiereSuchliste(suchListe);
					for (final HierarchieObjekt eintrag : suchListe) {
						hierarchieTypListe.add(eintrag.getObjekt().getType());
						final ParameterInfo hierarchieInfo = new ParameterInfo(eintrag.getObjekt(), info.getTyp(),
								info.getAtg(), info.getSim());
						parameter.setDataWithTime(parametrierung.getDefaultParameterDaten(hierarchieInfo));
						if (parameter.getData() != null) {
							parameter.setQuelle(eintrag.getObjekt(), QuellTyp.DEFAULT_HIERARCHIE);
							break;
						}
					}
				} while (parameter.getData() == null && suchListe.size() > 0);
			}
		}

		// gegebenenfalls Typdefinitionen durchsuchen
		if (parameter.getData() == null) {
			SystemObjectType type = info.getObjekt().getType();
			while (parameter.getData() == null && type != null) {
				final ParameterInfo typInfo = new ParameterInfo(type, info.getTyp(), info.getAtg(), info.getSim());
				parameter.setDataWithTime(parametrierung.getDefaultParameterDaten(typInfo));
				if (parameter.getData() != null) {
					parameter.setQuelle(type, QuellTyp.DEFAULT_TYP);
				}
				type = getValidSuperType(type, info.getAtg());
			}

			if (parameter.getData() == null) {
				for (final SystemObject obj : hierarchieTypListe) {
					final ParameterInfo typInfo = new ParameterInfo(obj, info.getTyp(), info.getAtg(), info.getSim());
					parameter.setDataWithTime(parametrierung.getDefaultParameterDaten(typInfo));
					if (parameter.getData() != null) {
						break;
					}
				}
			}
		}

		return parameter;
	}

	/**
	 * expandiert die Suchliste bei der Suche innerhalb der Objekthierachie.<br>
	 * Alle in der übergebenen Liste enthaltenen Objekte werden durch ihre
	 * Vorgäangerobjekte ersetzt.
	 *
	 * @param suchListe
	 *            die zu expandierende Liste
	 * @return die expandierte Liste
	 */
	private Collection<HierarchieObjekt> expandiereSuchliste(final Collection<HierarchieObjekt> suchListe) {
		final Collection<HierarchieObjekt> result = new ArrayList<>();

		for (final HierarchieObjekt objekt : suchListe) {
			result.addAll(objekt.getVorgaengerListe());
		}
		return result;
	}

	/**
	 * ermittelt einen gültigen Supertyp für den übergebenen Typ.<br>
	 * Ein Supertyp ist gültig, wenn er die übergebene Attributgruppe enthält.
	 *
	 * @param type
	 *            der Typ
	 * @param atg
	 *            die Attributgruppe
	 * @return den ermitteltn Typ oder <i>null</i>, wenn kein solcher existiert
	 */
	private SystemObjectType getValidSuperType(final SystemObjectType type, final AttributeGroup atg) {
		SystemObjectType result = null;
		final Collection<SystemObjectType> baseTypes = type.getSuperTypes();
		for (final SystemObjectType eintrag : baseTypes) {
			if (eintrag.getAttributeGroups().contains(atg)) {
				result = eintrag;
				break;
			}
		}

		return result;
	}
}
